#ifndef _TAGIO_HPP
#define _TAGIO_HPP
#include <string>
#include <sstream>
#include <vector>
#include <map>
#include <algorithm>
#include <iomanip>
#include <sys/ioctl.h>
#include <sys/sysinfo.h>
#include <typeinfo>
#include <iodb/iodb.h>

using namespace std;

/* Enumerations */

/*!
 * \brief The DataTypeStates enum Enumeration of each state in the data type state table
 */
enum DataTypeStates
{
    START,    // 0
    OFFSET1,  // 1
    OFFSET2,  // 2
    TYPE,     // 3
    FIRST,    // 4
    SECOND,   // 5
    THIRD,    // 6
    OPTIONAL, // 7
    P_SIZE,   // 8
    DEFAULT   // 9 (No State)
};

/*!
 * \brief The TagType enum Enumeration for tag types.
 *
 * Each tag type determines the kind of value that is read from or written to the tag.
 */
enum TagType
{
    TAG_BIT,
    TAG_16,
    TAG_U16,
    TAG_32,
    TAG_U32,
    TAG_64,
    TAG_U64,
    TAG_FLOAT,
    TAG_DOUBLE,
    TAG_STRING,
    TAG_ERROR
};

/*!
 * \brief The CharacterKey enum The enumeration for the single letter character key (I, F, B, A)
 */
enum CharacterKey
{
    CK_UNKNOWN,
    CK_FLOAT, // F
    CK_INTEGER, // I
    CK_ASCII, // A
    CK_BIT // B
};

/*!
 * \brief The AdditionalInterpretation enum Enumeration for the additional optional data type information
 */
enum AdditionalInterpretation
{
    AD_SIGNED,
    AD_UNSIGNED,
    AD_BIG_ENDIAN,
    AD_LITTLE_ENDIAN,
    AD_PRECISION
};

/*!
 * \brief The TagErrors enum Contains all the error codes of the api
 */
enum TagErrors
{
    // TE = Tag Error
    TE_NO_ERROR, // No error
    TE_NOT_NUMERIC, // Value to be written to IODB is not numeric
    TE_READ_ERROR,  // Failed to read from IODB register
    TE_WRITE_ERROR, // Failed to write to IODB register
    TE_READ_ONLY,   // IODB register is read only
    TE_DISCRETE_ONLY, // IODB register accepts only 0 or 1
    TE_ANALOG_ONLY, // IODB register accepts only 0 to 65535 (16 bit)
    TE_FLOAT_ONLY, // IODB register accepts only floats (decimal values)
    TE_LONG_ONLY, // IODB register accepts only 64 bit values
    TE_VALUE_TOO_LARGE, // Value is too large for its data type
    TE_VALUE_IS_SIGNED, // Value received is signed, expected unsigned value
    TE_VALUE_IS_UNSIGNED ,// Value received is unsigned, expected signed value
    TE_INVALID_ENDIAN, // Invalid endian value passed
    TE_INVALID_CHARACTER_KEY, // Invalid character key passed
    TE_INVALID_ADDITIONAL, // Invalid additional information
    TE_INVALID_CHAR, // Invalid character for parsing Data Type Field
    TE_INVALID_IODB_TYPE, // Invalid IODB Type number
    TE_INVALID_OFFSET, // Invalid offset value for given bitsize
    TE_INVALID_BITSIZE, // Invalid bitsize, accepted 16, 32, 64
    TE_INVALID_PRECISION, // Invalid percision value, accepted 0-9
    TE_PRECISION_NOT_ALLOWED, // Precicion not allowed for the data type
    TE_SIGNED_NOT_ALLOWED, // The tag type does not allow for signed values
    TE_UNSIGNED_NOT_ALLOWED, // The tag type does not allow for unsigned values
    TE_INVALID_CHAR_LENGTH, // Invalid length for ASCII Data Type. Vlaid 1-255
    TE_INVALID_DATA_TYPE_LENGTH, // Data type string is too large,  Max length 30 characters
    TE_INVALID_REGISTER_ADDRESS, // The register address is invalid
    TE_TAG_ALREADY_EXISTS, // The tag already exists
    TE_IODB_TYPE_MISSING, // The iodb type is missing
    TE_TAG_NOT_FOUND, // The tag was not found
    TE_FILE_DOES_NOT_EXIST, // The file does not exist
    TE_CORRUPT_CSV_FILE, // Corrupted csv file, or invalid csv format
    TE_DATA_TYPE_MISSING, // The data type field is missing
    TE_INVALID_TAG_TYPE, // The tag type is not valid
    TE_INVALID_INDEX, // The index was invalid
    TE_EMPTY_STRING, // The string passed was empty
    TE_INVALID_NUMERIC_TYPE, // The numerical type does not match the iodb type
    TE_DUMMY_TAG_CREATED // A dummy tag was created
};

/*!
 * \brief The TagColumn enum Enumeration of the columns in the tags.csv file
 */
enum TagColumn
{
    TAG_COL_TAG_NAME,
    TAG_COL_STATION_NAME,
    TAG_COL_STATION_NUMBER,
    TAG_COL_IO_TYPE_NUMBER,
    TAG_COL_IO_REGISTER_NUMBER,
    TAG_COL_DATA_TYPE,
    TAG_COL_ACTUAL_MIN,
    TAG_COL_ACTUAL_MAX,
    TAG_COL_SCALED_MIN,
    TAG_COL_SCALED_MAX,
    TAG_COL_ENGINEERING_UNITS,
    TAG_COL_DECIMALS,
    TAG_COL_OFF_MESSAGE,
    TAG_COL_ON_MESSAGE,
    TAG_COL_DESCRIPTION,
    TAG_COL_RETAIN,
    TAG_COL_SOURCE,
    TAG_COL_CLASS,
    TAG_COL_DEADBAND,
    TAG_COL_COUNT
};

/*!
 * \brief The IODBType enum Enumeration of the different IODB types.
 */
enum IODBType
{
    ANALOG_INPUT = 0,
    ANALOG_OUTPUT = 1,
    DISCRETE_INPUT = 10,
    DISCRETE_OUTPUT = 11,
    LONG_INPUT = 20,
    LONG_OUTPUT = 21,
    FLOAT_INPUT = 22,
    FLOAT_OUTPUT = 23
};

/* Data Structure */
/*!
 * \brief The IODBTag struct The main data structure for the IODB Tag.
 */
struct IODBTag
{
    // Description
    string name;
    /// IODB type
    IODBType    iodbType;
    /// IODB register address
    int    iodbRegister;
    /// Type of value based on TagType enusm
    int    tagType;

    // Current Values
    /// Descrite, bit or a Flag value
    bool   bitValue;
    /// Short value of the IODB
    short    value16;
    /// Unsigned short value of the IODB
    unsigned short uValue16;
    /// Integer value of the IODB
    long      value32;
    /// Unsigned integer value of the IODB
    unsigned long  uValue32;
    /// Long integer value
    long long   value64;
    /// Unsigned Long integer value
    unsigned long long   uValue64;
    /// Float value of the IODB
    float  floatValue;
    /// Double value of the IODB
    double doubleValue;
    /// String value of the tag
    string stringValue;

    // Previously Read Values
    /// Previously Read Descrite, bit or a Flag value
    bool   prevBitValue;
    /// Previously Read Short value of the IODB
    short    prevValue16;
    /// Previously Read Unsigned short value of the IODB
    unsigned short prevUValue16;
    /// Previously Read Integer value of the IODB
    long      prevValue32;
    /// Previously Read Unsigned integer value of the IODB
    unsigned long  prevUValue32;
    /// Previously Read Long integer value
    long long   prevValue64;
    /// Previously Read Unsigned Long integer value
    unsigned long long   prevUValue64;
    /// Previously Read Float value of the IODB
    float  prevFloatValue;
    /// Previously Read Double value of the IODB
    double prevDoubleValue;
    /// Previously Read String value of the tag
    string prevStringValue;

    // Data Type Section
    /// Data Type value
    string    dataType;
    int offset;
    int characterKey;
    int bitSize;
    int sign;
    int endian;
    int precision;

    // Additional Information
    /// Flag if the IODB is read only
    bool   isReadOnly;
    /// String containing the unit type of the register
    string units;
    /// Range in which the value is within the previous reported value
    short deadband;
    float floatDeadband;
    long  longDeadband;

    // Inspector Functions

    bool getBitValue()
    {
        return bitValue;
    }

    short getValue16()
    {
        return value16;
    }

    unsigned short getUValue16()
    {
        return uValue16;
    }

    long getValue32()
    {
        return value32;
    }

    unsigned long getUValue32()
    {
        return uValue32;
    }

    long long getValue64()
    {
        return value64;
    }

    unsigned long long getUValue64()
    {
        return uValue64;
    }

    float getFloatValue()
    {
        return floatValue;
    }

    string getStringValue()
    {
        return stringValue;
    }

    string getUnits()
    {
        return units;
    }

    template <class T>
    IODBTag& operator = (T value)
    {
        switch(tagType)
        {
        case TAG_BIT:
            bitValue = value;
            break;
        case TAG_16:
            value16 = value;
            break;
        case TAG_U16:
            uValue16 = value;
            break;
        case TAG_32:
            value32 = value;
            break;
        case TAG_U32:
            uValue32 = value;
            break;
        case TAG_64:
            value64 = value;
            break;
        case TAG_U64:
            uValue64 = value;
            break;
        case TAG_FLOAT:
            floatValue = value;
            break;
        case TAG_DOUBLE:
            doubleValue = value;
            break;
        default:
            // Do nothing
            break;
        }

        return *this;
    }
};

/* Template Functions */

template <class T>
/*!
 * \brief toString Template function to convert the value to a string
 * \param t The value to be converted
 * \return The value as a string
 */
inline string toString(const T& t)
{
    stringstream ss;
    ss << t;
    return ss.str();
}

template <class T>
/*!
 * \brief floatToString Template function to convert the value to a string with precision set to 6
 * \param t The value to be converted
 * \return The value as a string
 */
inline string floatToString(const T& t)
{
    stringstream ss;
    ss.precision(6);

    if(0 == t)
    {
        ss << t;
    }
    else if(fabs(t) > 1.0e9 || fabs(t) < 1.0e-6)
    {
        ss << scientific << t;
    }
    else
    {
         ss << fixed << t;
    }
    return ss.str();
}

template <typename T>
/*!
 * \brief absoluteValue Returns the absolute value of the value passed
 * \param value The value to be calculated
 * \return The absolute value of the value
 */
T absoluteValue(T value)
{
    return (value > 0) ? value : -value;
}

template <typename T>
/*!
 * \brief diff Takes two values and compares them for the larger value. Then subtracts the smaller from the larger
 *
 * This function is mainly used to get a difference in the form of a positive instead of negative. This is useful
 * when trying to get a difference from two unsigned numbers that cannot represent a true signed negative number.
 * \param a The first number
 * \param b The second number
 * \return The positive difference between the two numbers
 */
T diff(T a, T b)
{
   return (a > b) ? (a - b) : (b - a);
}

template <class T>
/*!
 * \brief hasValueChanged Checks if the difference between "currentValue" and "oldValue" is greater than "deadband"
 * \param currentValue The current value
 * \param oldValue The old value
 * \param deadband The difference threshold to exceed to register a change
 * \return True is change is recognized, other wise false.
 */
bool hasValueChanged(T currentValue, T &oldValue, T deadband)
{
    if(absoluteValue(diff(currentValue, oldValue)) > deadband)
    {
        // If the difference between the current value and the base value exceeds the deadband threshold
        // then the current value becomes the new base value
        oldValue = currentValue;
        return true;
    }
    else
    {
        return false;
    }
}

template <class T>
/*!
 * \brief isTagValue Checks whether the current value in the tag matches the passed value
 * \param tag The IODBTag to check
 * \param value The value to compare with the tag's current value
 * \return True if the value matches, else false
 */
bool isTagValue(IODBTag tag, T value)
{
    switch(tag.tagType)
    {
    case TAG_BIT:
        return tag.bitValue == value;
        break;
    case TAG_16:
        return tag.value16 == value;
        break;
    case TAG_U16 :
        return tag.uValue16 == value;
        break;
    case TAG_32:
        return tag.value32  == value;
        break;
    case TAG_U32:
        return tag.uValue32 == value;
        break;
    case TAG_64:
        return tag.value64 == value;
        break;
    case TAG_U64:
        return tag.uValue64 == value;
        break;
    case TAG_FLOAT :
        return tag.floatValue == value;
        break;
    case TAG_DOUBLE:
        return tag.doubleValue == value;
        break;
    default:
        return false;
    }
}

/* Search Predicates */

/*!
 * \brief The MatchesName class Search predicate for finding a tag by name case sensitive
 */
class MatchesName
{
    std::string _name;

    public:
        MatchesName(const string &name) : _name(name) {}

        bool operator()(const IODBTag &item) const
        {
            return item.name == _name;
        }
};

/*!
 * \brief The MatchesNameCaseInsensitive class Search predicate for finding a tag by name case insensitive
 */
class MatchesNameCaseInsensitive
{
    std::string _name;

    public:
        MatchesNameCaseInsensitive(const string &name) : _name(name) {}

        bool operator()(const IODBTag &item) const
        {
            string tagName = item.name;
            string queryString = _name;
            transform(tagName.begin(),tagName.end(),tagName.begin(),::tolower);
            transform(queryString.begin(),queryString.end(),queryString.begin(),::tolower);
            return tagName == queryString;
        }
};

/*!
 * \brief The MatchesIODBRegister class Search predicate for finding a tag by iodb type and reigster
 */
class MatchesIODBRegister {

    int _iodbRegister;
    IODBType _iodbType;

    public:
        MatchesIODBRegister(const IODBType iodbType, const int iodbRegister) :
            _iodbType(iodbType), _iodbRegister(iodbRegister) {}

        bool operator()(const std::pair<string, IODBTag> &item) const
        {
            return (item.second.iodbRegister == _iodbRegister &&
                item.second.iodbType == _iodbType);
        }
};

/* Global Non Class Specific Functions */

// Data Conversion
int convert32to64(unsigned int high, unsigned int low, int endian, unsigned long long &result);
int convert16to32(unsigned short high, unsigned short low, int endian, unsigned long &result);
int convert16to64(unsigned short first, unsigned short second, unsigned short third, unsigned short fourth, int endian, unsigned long long &result);
int convert64to16(unsigned long long U64BitNumber, int endian, unsigned short *U16BitParts);
int convert32to16(unsigned long U32BitNumber, int endian, unsigned short *U16BitParts);
int convert16toFloat(unsigned short high, unsigned short low, int endian, float &result);
int convert16toDouble(unsigned short first, unsigned short second, unsigned short third, unsigned short fourth, int endian, double &result);
int convertFloatTo16(float floatValue, int endian, unsigned short *U16BitParts);
int convertDoubleTo16(double doubleValue, int endian, unsigned short *U16BitParts);

void setNthBit(short &value, int set, int n);

// Type Checking And Comparison
bool containsOnlyDigits(string value);
bool isNumericValue(const char *stringValue);
bool hasStringValueChanged(string currentValue, string &oldValue);

// Enumeration Decoding and Encoding
int getEnumeratedIodbType(string iodbType, IODBType &value);
string getIODB_Type(int iodbType);
int inputToIndex(char input, int &rc);
string errorToString(TagErrors error);
int parseCharacterKey(string characterKey, IODBTag &tag);
// Parses the character key property of the tag data type string
int parseAdditional(string additional, IODBTag &tag);

// Returns the default value of a dummy tag.
IODBTag initializeDefaultTag();

/*!
 * \brief The TagIO class The main api class that holds all the methods needed for Tag I/O operations
 */
class TagIO
{
public:
    // Local Return Code for internal use if needed
    int rc;
    /*************************/
    /* Local Data Structures */
    /*************************/

    // A tag name to index map for use when checking for existance of a Tag or for faster random
    // access to a tag using the Tag's name. This particular map handles the Tags that TagIO
    // knows exist. These tags represent the Tags that exist in the tags.csv file.
    map<string, IODBTag> tagsMap;

    // A tag name to index map for use when checking for existance of a Tag or for faster random
    // access to a tag using the Tag's name. This particular map handles the Tags that the user
    // program will actually use.
    map<string, IODBTag> userListMap;

    /**************************************/
    /* Class Constructors and Destructors */
    /**************************************/

    // Default constructor. Will populate the "tags" list by reading from a system defined csv file
    TagIO();

    // Constructor that will also populate the "tags" list by reading from a specific csv file
    TagIO(string tagsCSVFile);

    /**********************/
    /* User Api Functions */
    /**********************/

    /* Tag List Population */

    // Populates the "userList" list by interpreting a formatted string. Ex. "AO:3000,AO:3001,..."
    int addTagsFromString(string tagList);

    // Adds all the tags from the tags list to the userList
    int addAllTags();

    // Adds a single tag to the userList. Tag is referenced by name.
    int addTag(string name, bool caseSensitive = false);

    // Adds a single tag to the userList. Tag is reference by iodb type and address.
    int addTag(IODBType iodbType, int address);

    /* Tag Evaluation Functions */

    // Overloaded function. Checks if a tag value has changed. Tag is referenced by name.
    bool hasTagValueChanged(string tagName, bool caseSensitive = false);

    // Overloaded function. Checks if a tag value has changed. Tag is reference by iodb type and address.
    bool hasTagValueChanged(IODBType iodbType, int address);

    // Overloaded function. Checks if a tag value has changed. Tag is referenced by reference
    bool hasTagValueChanged(IODBTag &tag);

    // Checks if a float value is within a valid range
    bool isValidFloat(float floatValue);

    // Checks if a double value is within a valid range
    bool isValidDouble(double doubleValue);


    /* Iteration Helpers */

    // Returns the size of the user list
    int getUserListSize();

    // Returns the tag by value. Tag is referenced by name.
    IODBTag getTag(string tagName, bool caseSensitive = false);

    // Returns the tag by value. Tag is referenced by iodb type and address
    IODBTag getTag(IODBType iodbType, int address);

    // Returns the tag by iterator. Tag is referenced by name.
    map<string, IODBTag>::iterator getTagIter(string tagName);

    /* Updating Tag Values */

    // Updates the tag values in the "userList"
    int updateTags();

    /* Reading and Writing Tag Values */

    // Overloaded function for writing a value represented as a string to a tag. Tag is referenced by name.
    int writeTag(string name, string value, bool caseSensitive = false);
    // Overloaded function for writing a value represented as a string to a tag. Tag is referenced by iodb type and address.
    int writeTag(IODBType iodbType, int address, string value);
    int writeTag(map<string, IODBTag>::iterator &it, string value);

    int writeTag(string name, char value);
    int writeTag(int iodbType, int address, char value);
    int writeTag(map<string, IODBTag>::iterator &it, char value);

    int writeTag(string name, const char* value, bool caseSensitive);
    int writeTag(IODBType iodbType, int address, const char* value);
    int writeTag(map<string, IODBTag>::iterator &it, const char *value);

    template <class T>
    /*!
     * \brief writeTag Templated function to write a value to a partiuclar tag based on the tag name
     * \param name Name of the tag
     * \param value Value to write to the tag
     * \param caseSensitive If true, then case will be ignored when finding a tag name with the matching string
     * \return TE_NO_ERROR on success
     *         TE_TAG_NOT_FOUND if the tag was not found
     *         TE_INVALID_TAG_TYPE if the tag type for the tag is invalid
     */
    int writeTag(string name, T value, bool caseSensitive = false)
    {
        int rc = TE_NO_ERROR;

        map<string, IODBTag>::iterator it;
        if (TE_NO_ERROR != findTag(name, userListMap, it, caseSensitive))
        {
            rc = TE_TAG_NOT_FOUND;
            return rc;
        }

        rc = writeTag(it, value);
        return rc;
    }

    // Template function for writing a value to a tag. Tag is referenced by iodb type and address.
    template <class T>
    /*!
     * \brief writeTag Templated function to write a value to a partiuclar tag based on iodbtype and address
     * \param iodbType Enumerated iodbType
     * \param address The iodb register address
     * \param value The value to write to the tag
     * \return TE_NO_ERROR on success
     *         TE_TAG_NOT_FOUND if the tag was not found
     *         TE_INVALID_TAG_TYPE if the tag type for the tag is invalid
     */
    int writeTag(IODBType iodbType, int address, T value)
    {
        int rc = TE_NO_ERROR;

        map<string, IODBTag>::iterator it;
        if (TE_NO_ERROR != findTag(iodbType, address, userListMap, it))
        {
            rc = TE_TAG_NOT_FOUND;
            return rc;
        }

        rc = writeTag(it, value);
        return rc;
    }

    // Retrieves the latest value of a tag. Tag is referenced by index
    string getTagValueAsString(IODBTag tag);
    // Retrieves the latest value of a tag. Tag is referenced by name
    string getTagValueAsString(string tagName, bool caseSensitive = false);
    // Retrieves the latest value of a tag. Tag is referenced by iodb type and address
    string getTagValueAsString(IODBType iodbType, int address);
    string getFloatStringValue(IODBTag & tag);

    /* Retrieving Tag Attributes */

    // Retrives the tag's units. Tag is referenced by index
    string getTagUnits(string tagName);


    /*************************************/
    /* Non User Called Utility Functions */
    /*************************************/

    /* Tag List Population */

    // Reads a specified tag csv file
    int readTagCSVFile(string filePath);

    // Loads tags interpreted from a specified csv file to the "tags" tag list
    int loadTagsFromFile(string tagsCSVFile);

    // Adds a single tag that has been parsed from a csv file
    int addTagToTagsList(IODBTag tag);

    // Adds a single tag to the user list
    int addTagToUserList(IODBTag tag);

    // Populate tag initial values from IODB
    int setTagInitialValues(IODBTag &tag);

    // Clear userList and index map
    void clearUserTags();
    /* Tag Initialization */

    // Assigns the default values of a tag based on an iodb type and address to the tag provided
    int initializeTag(int iodbType, int address, IODBTag &tag);

    /* Tag Creation and Parsing  */

    // Set the tag type of a tag based on the current tag properties
    int setTagType(IODBTag &tag);

    // Parses the formatted data type string of the tag and populates data type properties
    int parseDataType(IODBTag &tag);

    /* Tag Validation */
    int validateDataType(IODBTag &tag);

    /* Tag Evaluation */


    /* Tag Searching and Matching */

    // Finds a tag by name
    int findTag(string tagName, map<string, IODBTag> &tagList, map<string, IODBTag>::iterator &it, bool caseSensitive = false);

    // Finds a tag by iodb type and address
    int findTag(IODBType iodbType, int address, map<string, IODBTag> &tagList, map<string, IODBTag>::iterator &it);

    /* Tag Reading and Updating */

    // Updates the tag provided
    int readTag(struct IODBTag &tag);

    // Remove Tag
    int removeTag(string name, bool caseSensitive = false);
    int removeTag(IODBType iodbType, int address);

    /* Tag Writing */

    template <class T>
    int writeTag(map<string, IODBTag>::iterator &it, T value)
    {
        int rc = TE_NO_ERROR;

        if(it->second.isReadOnly)
        {
            rc = TE_READ_ONLY;
            return rc;
        }

        // Holds parts of a larger number for instance where the tag's bitsize is larger than 16 bits
        unsigned short U16BitParts[4] = {0,0,0,0};
        //What is our offset so we know which iobb register to start with
        int startingRegister = it->second.iodbRegister - it->second.offset;

        switch(it->second.tagType)
        {
        case TAG_BIT:
            // Consider offset in register and assign that bit
            IODBWrite(it->second.iodbType,it->second.iodbRegister, 1, (void*)&value, NULL);
            break;
        case TAG_16:
        case TAG_U16:
            IODBWrite(it->second.iodbType, it->second.iodbRegister, 1, (void*)&value, NULL);
            break;
        case TAG_32:
            convert32to16((unsigned long)value, it->second.endian, U16BitParts);
            IODBWrite(it->second.iodbType, startingRegister, 2, (void*)&U16BitParts, NULL);
            break;
        case TAG_U32:
            // Consider the offset and split value into two 16 bit unsigned numbers
            convert32to16((unsigned long)value, it->second.endian, U16BitParts);
            IODBWrite(it->second.iodbType, startingRegister, 2, (void*)&U16BitParts, NULL);
            break;
        case TAG_64:
            // Consider the offset and split value into two 16 bit unsigned numbers
            convert64to16((unsigned long long)value, it->second.endian, U16BitParts);
            IODBWrite(it->second.iodbType, startingRegister, 4, (void*)&U16BitParts, NULL);
            // Consider the offset and split value into 4 16 bit signed numbers
            break;
        case TAG_U64:
            // Consider the offset and split value into two 16 bit unsigned numbers
            convert64to16((unsigned long long)value, it->second.endian, U16BitParts);
//            for(int i = 0; i < 4; i++)
//            {
//                cout << "Write Part[" << i << "] is " << U16BitParts[i] << endl;
//            }
            IODBWrite(it->second.iodbType, startingRegister, 4, (void*)&U16BitParts, NULL);
            // Consider the offset and split value into 4 16 bit unsigned numbers
            break;
        case TAG_FLOAT:
            if(FLOAT_INPUT == it->second.iodbType || FLOAT_OUTPUT == it->second.iodbType)
            {
                IODBWrite(it->second.iodbType, startingRegister, 1, (void*)&value, NULL);
            }
            else
            {
                convertFloatTo16(value, it->second.endian, U16BitParts);
                IODBWrite(it->second.iodbType, startingRegister, 2, (void*)&U16BitParts, NULL);
            }
            break;
        case TAG_DOUBLE:
            convertDoubleTo16(value, it->second.endian, U16BitParts);
            IODBWrite(it->second.iodbType, startingRegister, 4, (void*)&U16BitParts, NULL);
            break;
        case TAG_STRING:
            // Consider how many IODB Registers are need and split string into 2 char segments
            // for each underlying iodb register.
            /*
            IODBWrite(it->second.iodbType, startingRegister, 1, (void*)&value, NULL);
            */
            break;
        default:
            rc = TE_INVALID_TAG_TYPE;
        }

        return rc;
    }
};

/* Operator Overloads */
template <class T>
inline T operator + (IODBTag tag, T value)
{
    switch(tag.tagType)
    {
    case TAG_BIT:
        return tag.bitValue + value;
        break;
    case TAG_16:
        return tag.value16 + value;
        break;
    case TAG_U16:
        return tag.uValue16 + value;
        break;
    case TAG_32:
        return tag.value32  + value;
        break;
    case TAG_U32:
        return tag.uValue32 + value;
        break;
    case TAG_64:
        return tag.value64 + value;
        break;
    case TAG_U64:
        return tag.uValue64 + value;
        break;
    case TAG_FLOAT:
        return tag.floatValue + value;
        break;
    case TAG_DOUBLE:
        return tag.doubleValue + value;
        break;
    default:
        return 0;
    }
}

template <class T>
inline T operator - (IODBTag tag, T value)
{
    switch(tag.tagType)
    {
    case TAG_BIT:
        return tag.bitValue - value;
        break;
    case TAG_16:
        return tag.value16 - value;
        break;
    case TAG_U16:
        return tag.uValue16 - value;
        break;
    case TAG_32:
        return tag.value32 - value;
        break;
    case TAG_U32:
        return tag.uValue32 - value;
        break;
    case TAG_64:
        return tag.value64 - value;
        break;
    case TAG_U64:
        return tag.uValue64 - value;
        break;
    case TAG_FLOAT:
        return tag.floatValue - value;
        break;
    case TAG_DOUBLE:
        return tag.doubleValue - value;
        break;
    default:
        return 0;
    }
}

template <class T>
inline bool operator > (IODBTag tag, T value)
{
    switch(tag.tagType)
    {
    case TAG_BIT:
        return tag.bitValue > value;
        break;
    case TAG_16:
        return tag.value16 > value;
        break;
    case TAG_U16:
        return tag.uValue16 > value;
        break;
    case TAG_32:
        return tag.value32  > value;
        break;
    case TAG_U32:
        return tag.uValue32 > value;
        break;
    case TAG_64:
        return tag.value64 > value;
        break;
    case TAG_U64:
        return tag.uValue64 > value;
        break;
    case TAG_FLOAT:
        return tag.floatValue > value;
        break;
    case TAG_DOUBLE:
        return tag.doubleValue > value;
        break;
    default:
        return false;
    }
}

template <class T>
inline bool operator >= (IODBTag tag, T value)
{
    switch(tag.tagType)
    {
    case TAG_BIT:
        return tag.bitValue >= value;
        break;
    case TAG_16:
        return tag.value16 >= value;
        break;
    case TAG_U16:
        return tag.uValue16 >= value;
        break;
    case TAG_32:
        return tag.value32  >= value;
        break;
    case TAG_U32:
        return tag.uValue32 >= value;
        break;
    case TAG_64:
        return tag.value64 >= value;
        break;
    case TAG_U64:
        return tag.uValue64 >= value;
        break;
    case TAG_FLOAT:
        return tag.floatValue >= value;
        break;
    case TAG_DOUBLE:
        return tag.doubleValue >= value;
        break;
    default:
        return false;
    }
}

template <class T>
inline bool operator < (IODBTag tag, T value)
{
    switch(tag.tagType)
    {
    case TAG_BIT:
        return tag.bitValue < value;
        break;
    case TAG_16:
        return tag.value16 < value;
        break;
    case TAG_U16:
        return tag.uValue16 < value;
        break;
    case TAG_32:
        return tag.value32  < value;
        break;
    case TAG_U32:
        return tag.uValue32 < value;
        break;
    case TAG_64:
        return tag.value64  < value;
        break;
    case TAG_U64:
        return tag.uValue64 < value;
        break;
    case TAG_FLOAT:
        return tag.floatValue < value;
        break;
    case TAG_DOUBLE:
        return tag.doubleValue < value;
        break;
    default:
        return false;
    }
}

template <class T>
inline bool operator <= (IODBTag tag, T value)
{
    switch(tag.tagType)
    {
    case TAG_BIT:
        return tag.bitValue <= value;
        break;
    case TAG_16:
        return tag.value16 <= value;
        break;
    case TAG_U16:
        return tag.uValue16 <= value;
        break;
    case TAG_32:
        return tag.value32  <= value;
        break;
    case TAG_U32:
        return tag.uValue32 <= value;
        break;
    case TAG_64:
        return tag.value64  <= value;
        break;
    case TAG_U64:
        return tag.uValue64 <= value;
        break;
    case TAG_FLOAT:
        return tag.floatValue <= value;
        break;
    case TAG_DOUBLE:
        return tag.doubleValue <= value;
        break;
    default:
        return false;
    }
}

template <class T>
inline bool operator == (IODBTag tag, T value)
{
    switch(tag.tagType)
    {
    case TAG_BIT:
        return tag.bitValue == value;
        break;
    case TAG_16:
        return tag.value16 == value;
        break;
    case TAG_U16:
        return tag.uValue16 == value;
        break;
    case TAG_32:
        return tag.value32  == value;
        break;
    case TAG_U32:
        return tag.uValue32 == value;
        break;
    case TAG_64:
        return tag.value64  == value;
        break;
    case TAG_U64:
        return tag.uValue64 == value;
        break;
    case TAG_FLOAT:
        return tag.floatValue == value;
        break;
    case TAG_DOUBLE:
        return tag.doubleValue == value;
        break;
    default:
        return false;
    }
}

template <class T>
inline bool operator != (IODBTag tag, T value)
{
    switch(tag.tagType)
    {
    case TAG_BIT:
        return tag.bitValue != value;
        break;
    case TAG_16:
        return tag.value16 != value;
        break;
    case TAG_U16 :
        return tag.uValue16 != value;
        break;
    case TAG_32:
        return tag.value32  != value;
        break;
    case TAG_U32:
        return tag.uValue32 != value;
        break;
    case TAG_64:
        return tag.value64  != value;
        break;
    case TAG_U64:
        return tag.uValue64 != value;
        break;
    case TAG_FLOAT :
        return tag.floatValue != value;
        break;
    case TAG_DOUBLE:
        return tag.doubleValue != value;
        break;
    default:
        return false;
    }
}

/*!
 * \brief operator << Overloads the "<<" operator to make printing out the tag contents of the userList easier
 * \param stream << The output stream to append the tag list contents to
 * \param tagIO << The tagIO object containing the userList
 * \return The ostream reference for use with cout
 */
inline ostream& operator << (ostream& stream, TagIO* tagIO)
{
    map<string, IODBTag>::iterator it;
    for(it = tagIO->userListMap.begin(); it != tagIO->userListMap.end(); ++it)
    {
        stream << "Tag Name : " << it->second.name << endl;
        stream << "\t" << getIODB_Type(it->second.iodbType) << " " << toString(it->second.iodbRegister) << endl;
        stream << "\tOffset : " << it->second.offset << endl;
        stream << "\tCharacter Key : ";
        switch(it->second.characterKey)
        {
        case CK_ASCII:
            stream << "ASCII" << endl;
            break;
        case CK_BIT:
            stream << "BIT" << endl;
            break;
        case CK_FLOAT:
            stream << "FLOAT" << endl;
            break;
        case CK_INTEGER:
            stream << "INTEGER" << endl;
            break;
        default:
            stream << "INVALID CHARACTER KEY" << endl;
        }

        stream << "\tBit Size : " << it->second.bitSize << endl;
        stream << "\tSign : ";
        switch(it->second.sign)
        {
        case AD_SIGNED:
            stream << "SIGNED" << endl;
            break;
        case AD_UNSIGNED:
            stream << "UNSIGNED" << endl;
            break;
        default:
            stream << "INVALID SIGN" << endl;
            break;
        }
        stream << "\tEndian : ";
        switch(it->second.endian)
        {
        case AD_LITTLE_ENDIAN:
            stream << "LITTLE ENDIAN" << endl;
            break;
        case AD_BIG_ENDIAN:
            stream << "BIG ENDIAN" << endl;
            break;
        default:
            stream << "INVALID ENDIAN" << endl;
            break;
        }
        stream << "\tPrecision : " << it->second.precision << endl;
        stream << "\tIODB Enumerated Type : " << it->second.iodbType << endl;
        stream << "\tTag Type : ";

        switch(it->second.tagType)
        {
        case TAG_BIT:
            stream << "TAG_BIT" << endl;
            break;
        case TAG_16:
            stream << "TAG_16" << endl;
            break;
        case TAG_U16:
            stream << "TAG_U16" << endl;
            break;
        case TAG_32:
            stream << "TAG_32" << endl;
            break;
        case TAG_U32:
            stream << "TAG_U32" << endl;
            break;
        case TAG_64:
            stream << "TAG_64" << endl;
            break;
        case TAG_U64:
            stream << "TAG_U64" << endl;
            break;
        case TAG_FLOAT:
            stream << "TAG_FLOAT" << endl;
            break;
        case TAG_STRING:
            stream << "TAG_STRING" << endl;
            break;
        case TAG_DOUBLE:
            stream << "TAG_DOUBLE" << endl;
            break;
        default:
            stream << "Invalid Tag Type" << endl;
            break;
        }

        stream << "\tCurrent Value : ";
        switch(it->second.tagType)
        {
        case TAG_BIT:
            stream << it->second.bitValue << endl;
            break;
        case TAG_16:
            stream << it->second.value16 << endl;
            break;
        case TAG_U16:
            stream << it->second.uValue16 << endl;
            break;
        case TAG_32:
            stream << it->second.value32 << endl;
            break;
        case TAG_U32:
            stream << it->second.uValue32 << endl;
            break;
        case TAG_64:
            stream << it->second.value64 << endl;
            break;
        case TAG_U64:
            stream << it->second.uValue64 << endl;
            break;
        case TAG_FLOAT:
            stream << it->second.floatValue << endl;
            break;
        case TAG_STRING:
            stream << it->second.stringValue << endl;
            break;
        case TAG_DOUBLE:
            stream << fixed << setprecision(it->second.precision) << it->second.doubleValue << endl;
            break;
        default:
            stream << "Invalid Tag Type" << endl;
            break;
        }

        stream << "\tDeadband : ";
        switch(it->second.tagType)
        {
        case TAG_BIT:
        case TAG_STRING:
            stream << "Not Used" << endl;
            break;
        case TAG_16:
        case TAG_U16:
            stream << it->second.deadband << endl;
            break;
        case TAG_32:
        case TAG_U32:
        case TAG_64:
        case TAG_U64:
            stream << it->second.longDeadband << endl;
            break;
        case TAG_FLOAT:
        case TAG_DOUBLE:
            stream << it->second.floatDeadband << endl;
            break;
        default:
            stream << "Not Used" << endl;
            break;
        }
    }
    return stream;
}
#endif
